//+------------------------------------------------------------------+
//|                                                      Twitter.mqh |
//|                                            Copyright 2020, seffx |
//|                              https://www.mql5.com/en/users/seffx |
//+------------------------------------------------------------------+
//Published on MQL5 for Public Domain Usage

//#property copyright "Copyright 2020, seffx"
//#property link      "https://www.mql5.com/en/users/seffx"

#define TWITTER_REQUEST_TIMEOUT 5000
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
class CTwitter
  {
private:
   bool              m_verbose;
   string            m_consumer_key;
   string            m_consumer_secret;
   string            m_access_token;
   string            m_access_secret;

   //+------------------------------------------------------------------+
   //|                                                                  |
   //+------------------------------------------------------------------+
   int               SaveToFile(const string filename, const char &text[])
     {
      ResetLastError();
      int handle=FileOpen(filename,FILE_BIN|FILE_ANSI|FILE_WRITE);
      if(handle==INVALID_HANDLE)
         return(GetLastError());
      FileWriteArray(handle,text);
      FileClose(handle);
      return(0);
     }

   //+------------------------------------------------------------------+
   //|                                                                  |
   //+------------------------------------------------------------------+
   string            hex(int i)
     {
      static string h="0123456789ABCDEF";
      string ret="";
      int a = i % 16;
      int b = (i-a)/16;
      if(b>15)
         StringConcatenate(ret,ret,hex(b),StringSubstr(h,a,1));
      else
         StringConcatenate(ret,ret,StringSubstr(h,b,1),StringSubstr(h,a,1));
      return (ret);
     }
   //+------------------------------------------------------------------+
   //|                                                                  |
   //+------------------------------------------------------------------+
   string            URLEncode(string toCode)
     {
      int max=StringLen(toCode);

      string RetStr="";
      for(int i=0; i<max; i++)
        {
         string c = StringSubstr(toCode,i,1);
         ushort asc = StringGetCharacter(c, 0);

         if((asc >= '0' && asc <= '9')
            || (asc >= 'a' && asc <= 'z')
            || (asc >= 'A' && asc <= 'Z')
            || (asc == '-')
            || (asc == '.')
            || (asc == '_')
            || (asc == '~'))
            StringAdd(RetStr,c);
         else
           {
            StringConcatenate(RetStr,RetStr,"%",hex(asc));
           }
        }
      return (RetStr);
     }
   //+------------------------------------------------------------------+
   //|                                                                  |
   //+------------------------------------------------------------------+
   string            arrayEncode(string &array[][2])
     {
      string ret="";
      string key,val;
      int l=ArrayRange(array,0);
      for(int i=0; i<l; i++)
        {
         key = URLEncode(array[i,0]);
         val = URLEncode(array[i,1]);
         StringConcatenate(ret,ret,key,"=",val);
         if(i+1<l)
            StringConcatenate(ret,ret,"&");
        }
      return (ret);
     }
   //+------------------------------------------------------------------+
   //|                                                                  |
   //+------------------------------------------------------------------+
   void              sortParam(string&arr[][2])
     {
      string k1, k2;
      string v1, v2;
      int n = ArrayRange(arr,0);

      // bubble sort
      int i, j;
      for(i = 0; i < n-1; i++)
        {
         // Last i elements are already in place
         for(j = 0; j < n-i-1; j++)
           {
            int x = j+1;
            k1 = arr[j][0];
            k2 = arr[x][0];
            if(k1 > k2)
              {
               // swap values
               v1 = arr[j][1];
               v2 = arr[x][1];
               arr[j][1] = v2;
               arr[x][1] = v1;
               // swap keys
               arr[j][0] = k2;
               arr[x][0] = k1;
              }
            else
               if(k1 == k2)
                 {
                  // compare values
                  v1 = arr[j][1];
                  v2 = arr[x][1];
                  if(v1 > v2)
                    {
                     arr[j][1] = v2;
                     arr[x][1] = v1;
                     // swap keys
                     arr[j][0] = k2;
                     arr[x][0] = k1;
                    }
                 }
           }
        }
     }
   //+------------------------------------------------------------------+
   //|                                                                  |
   //+------------------------------------------------------------------+
   void              addParam(string key,string val,string&array[][2])
     {
      int x=ArrayRange(array,0);
      if(ArrayResize(array,x+1)>-1)
        {
         array[x][0]=key;
         array[x][1]=val;
        }
     }
   //+------------------------------------------------------------------+
   //|                                                                  |
   //+------------------------------------------------------------------+
   string            SendRequest(string method, string url, uchar &data[],
                                 string headers="", string params="")
     {
      char result[];
      string resp_headers;
      ResetLastError();
      StringToCharArray(params, data);
      ArrayResize(data, ArraySize(data)-1);
      int res = WebRequest(method, url, headers, TWITTER_REQUEST_TIMEOUT, data,
                           result, resp_headers);
      if(res != -1)
        {
         string resp = CharArrayToString(result);
         if(m_verbose)
           {
            Print("***");
            Print("Data:");
            Print(CharArrayToString(data));
            Print("Resp Headers:");
            Print(resp_headers);
            Print("Resp:");
            Print("***");
            Print(resp);
           }
         return resp;
        }
      else
        {
         int err = GetLastError();
         PrintFormat("* WebRequest error: %d (%d)", res, err);
         if(m_verbose)
           {
            Print("***");
            Print("Data:");
            Print(CharArrayToString(data));
           }
         if((err == 4014) || (err == 4060))
           {
            string msg = "* PLEASE allow https://api.twitter.com and "
                         "https://upload.twitter.com in WebRequest listed URL";
            Print(msg);
           }
        }
      return "";
     }
   //+------------------------------------------------------------------+
   //|                                                                  |
   //+------------------------------------------------------------------+
   string            getNonce()
     {
      const string alnum = "abcdef0123456789";
      char base[];
      StringToCharArray(alnum, base);
      int x, len = StringLen(alnum);
      char res[32];
      for(int i=0; i<32; i++)
        {
         x = MathRand() % len;
         res[i] = base[x];
        }
      return CharArrayToString(res);
     }
   //+------------------------------------------------------------------+
   //|                                                                  |
   //+------------------------------------------------------------------+
   string            getBase(string&params[][2], string url, string method="POST")
     {
      string s = method;
      StringAdd(s, "&");
      StringAdd(s, URLEncode(url));
      StringAdd(s, "&");
      bool first = true;
      int x=ArrayRange(params,0);
      for(int i=0; i<x; i++)
        {
         if(first)
            first = false;
         else
            StringAdd(s, "%26");  // URLEncode("&")
         StringAdd(s, URLEncode(params[i][0]));
         StringAdd(s, "%3D"); // URLEncode("=")
         StringAdd(s, URLEncode(params[i][1]));
        }
      return s;
     }
   //+------------------------------------------------------------------+
   //|                                                                  |
   //+------------------------------------------------------------------+
   string            getQuery(string&params[][2], string url = "")
     {
      string key;
      string s = url;
      string sep = "";
      if(StringLen(s) > 0)
        {
         if(StringFind(s, "?") < 0)
            sep = "?";
        }
      bool first = true;
      int x=ArrayRange(params,0);
      for(int i=0; i<x; i++)
        {
         key = params[i][0];
         if(StringFind(key, "oauth_")==0)
            continue;
         if(first)
           {
            first = false;
            StringAdd(s, sep);
           }
         else
            StringAdd(s, "&");
         StringAdd(s, params[i][0]);
         StringAdd(s, "=");
         StringAdd(s, params[i][1]);
        }
      return s;
     }
   //+------------------------------------------------------------------+
   //|                                                                  |
   //+------------------------------------------------------------------+
   string            getOauth(string&params[][2])
     {
      string key;
      string s = "OAuth ";
      bool first = true;
      int x=ArrayRange(params,0);
      for(int i=0; i<x; i++)
        {
         key = params[i][0];
         if(StringFind(key, "oauth_")!=0)
            continue;
         if(first)
            first = false;
         else
            StringAdd(s, ", ");
         StringAdd(s, URLEncode(key));
         StringAdd(s, "=\"");
         StringAdd(s, URLEncode(params[i][1]));
         StringAdd(s, "\"");
        }
      return s;
     }
   //+------------------------------------------------------------------+
   //|                                                                  |
   //+------------------------------------------------------------------+
   string            hmac_sha1(string smsg, string skey, uchar &dstbuf[])
     {
      // HMAC as described on:
      // https://en.wikipedia.org/wiki/HMAC
      //
      uint n;
      uint BLOCKSIZE=64;
      uchar key[];
      uchar msg[];
      uchar i_s[];
      uchar o_s[];
      uchar i_sha1[];
      uchar keybuf[];
      uchar i_key_pad[];
      uchar o_key_pad[];
      string s = "";

      if((uint)StringLen(skey)>BLOCKSIZE)
        {
         uchar tmpkey[];
         StringToCharArray(skey,tmpkey,0,StringLen(skey));
         CryptEncode(CRYPT_HASH_SHA1, tmpkey, keybuf, key);
         n=(uint)ArraySize(key);
        }
      else
         n=(uint)StringToCharArray(skey,key,0,StringLen(skey));

      if(n<BLOCKSIZE)
        {
         ArrayResize(key,BLOCKSIZE);
         ArrayFill(key,n,BLOCKSIZE-n,0);
        }

      ArrayCopy(i_key_pad,key);
      for(uint i=0; i<BLOCKSIZE; i++)
         i_key_pad[i]=key[i]^(uchar)0x36;

      ArrayCopy(o_key_pad,key);
      for(uint i=0; i<BLOCKSIZE; i++)
         o_key_pad[i]=key[i]^(uchar)0x5c;


      n=(uint)StringToCharArray(smsg,msg,0,StringLen(smsg));
      ArrayResize(i_s,BLOCKSIZE+n);
      ArrayCopy(i_s,i_key_pad);
      ArrayCopy(i_s,msg,BLOCKSIZE);

      CryptEncode(CRYPT_HASH_SHA1, i_s, keybuf, i_sha1);
      ArrayResize(o_s,BLOCKSIZE+ArraySize(i_sha1));
      ArrayCopy(o_s,o_key_pad);
      ArrayCopy(o_s,i_sha1,BLOCKSIZE);
      CryptEncode(CRYPT_HASH_SHA1, o_s, keybuf, dstbuf);

      for(int i=0; i < ArraySize(dstbuf); i++)
         StringConcatenate(s, s, StringFormat("%02x",(dstbuf[i])));

      return s;
     }

   //+------------------------------------------------------------------+
   //|                                                                  |
   //+------------------------------------------------------------------+
   string            getHeaders(string oauth, string custom_headers="",
                                string host = "api.twitter.com")
     {
      string headers = "Host:";
      StringAdd(headers, host);
      StringAdd(headers, "\r\nContent-Encoding: identity\r\nConnection: close\r\n");
      if(StringLen(custom_headers) > 0)
         StringAdd(headers, custom_headers);
      StringAdd(headers, "Authorization: ");
      StringAdd(headers, oauth);
      StringAdd(headers, "\r\n\r\n");
      return headers;
     }


   //+------------------------------------------------------------------+
   //|                                                                  |
   //+------------------------------------------------------------------+
   string            oauthRequest(string &params[][2], string url,
                                  string method="GET")
     {
      addParam("oauth_consumer_key", m_consumer_key, params);
      addParam("oauth_nonce", getNonce(), params);
      addParam("oauth_signature_method", "HMAC-SHA1", params);
      addParam("oauth_timestamp", IntegerToString(TimeGMT()), params);
      addParam("oauth_token", m_access_token, params);
      addParam("oauth_version", "1.0", params);
      sortParam(params);

      string query = getQuery(params, url);
      string base = getBase(params, url, method);
      uchar buf[];
      string key = URLEncode(m_consumer_secret);
      StringAdd(key, "&");
      StringAdd(key, URLEncode(m_access_secret));

      uchar hashbuf[], base64buf[], nokey[];
      string hash = hmac_sha1(base, key, hashbuf);
      CryptEncode(CRYPT_BASE64, hashbuf, nokey, base64buf);
      string base64 = CharArrayToString(base64buf);

      addParam("oauth_signature", base64, params);
      sortParam(params);
      return query;
     }

   //+------------------------------------------------------------------+
   //|                                                                  |
   //+------------------------------------------------------------------+
   void              appendPhoto(string filename, string hash, uchar &data[],
                                 bool common_flag=false)
     {
      int flags=FILE_READ|FILE_BIN|FILE_SHARE_WRITE|FILE_SHARE_READ;
      if(common_flag)
         flags|=FILE_COMMON;
      //---
      int handle=FileOpen(filename,flags);
      if(handle==INVALID_HANDLE)
         return;
      //---
      int size=(int)FileSize(handle);
      uchar img[];
      ArrayResize(img,size);
      FileReadArray(handle,img,0,size);
      FileClose(handle);
      int pos = ArraySize(data);
      int offset = pos + size;
      int hlen = StringLen(hash)+6;
      int newlen = offset + hlen;
      ArrayResize(data, newlen);
      ArrayCopy(data, img, pos);
      StringToCharArray("\r\n--"+hash+"\r\n", data, offset, hlen);
     }


public:

   //+------------------------------------------------------------------+
   //|                                                                  |
   //+------------------------------------------------------------------+
                     CTwitter(string _consumer_key, string _consumer_secret,
            string _access_token, string _access_secret, bool _verbose = false)
     {
      m_consumer_key = _consumer_key;
      m_consumer_secret = _consumer_secret;
      m_access_token = _access_token;
      m_access_secret = _access_secret;
      m_verbose = _verbose;
     }

   //+------------------------------------------------------------------+
   //|                                                                  |
   //+------------------------------------------------------------------+
   string            unquote(string str)
     {
      return StringSubstr(str, 1, StringLen(str) - 2);
     }
   //+------------------------------------------------------------------+
   //|                                                                  |
   //+------------------------------------------------------------------+
   string            getTokenValue(string msg, string token, string prefix=":",
                                   string postfix=",")
     {
      string pattern = "\"";
      StringAdd(pattern, token);
      StringAdd(pattern, "\"");
      StringAdd(pattern, prefix);
      int pos1 = StringFind(msg, pattern);
      if(pos1 < 0)
         return "";
      int pos2 = StringFind(msg, postfix, pos1);
      if(pos2 < 0)
         return "";
      pos1 = pos1 + StringLen(pattern);
      return(StringSubstr(msg, pos1, pos2 - pos1));
     }

   //+------------------------------------------------------------------+
   //|                                                                  |
   //+------------------------------------------------------------------+
   string              verifyCredentials()
     {
      string url = "https://api.twitter.com/1.1/account/verify_credentials.json";
      string params[][2];
      string query = oauthRequest(params, url);
      string o = getOauth(params);
      string headers = getHeaders(o);
      char data[];
      return SendRequest("GET", query, data, headers);
     }

   //+------------------------------------------------------------------+
   //|                                                                  |
   //+------------------------------------------------------------------+
   string              tweet(string msg, string medias="",
                             bool trimUser=false)
     {
      string url = "https://api.twitter.com/1.1/statuses/update.json";
      string params[][2];
      addParam("status", URLEncode(msg), params);
      if(StringLen(medias)>0)
         addParam("media_ids", URLEncode(medias), params);
      if(trimUser)
         addParam("trim_user", "1", params);
      string query = oauthRequest(params, url, "POST");
      string o = getOauth(params);
      string headers = getHeaders(o);
      char data[];
      return SendRequest("POST", query, data, headers);
     }

   //+------------------------------------------------------------------+
   //|                                                                  |
   //+------------------------------------------------------------------+
   string              uploadPhoto(string filename)
     {
      // POST multipart/form-data
      string url = "https://upload.twitter.com/1.1/media/upload.json";
      //string url = "https://httpbin.org/anything";
      string params[][2];
      string query = oauthRequest(params, url, "POST");
      string o = getOauth(params);
      string custom_headers = "Content-Type: multipart/form-data;"
                              " boundary=";
      string boundary = getNonce();
      StringAdd(custom_headers, boundary); // use nonce as boundary string
      StringAdd(custom_headers, "\r\n");
      string headers = getHeaders(o, custom_headers, "upload.twitter.com");

      //string query = getQuery(params, url);
      //Print(query);
      uchar data[];
      string part = "\r\n--";
      StringAdd(part, boundary);
      StringAdd(part, "\r\nContent-Disposition: form-data;"
                " name=\"media\"\r\n\r\n");
      StringToCharArray(part, data, 0, StringLen(part));
      appendPhoto(filename, boundary, data);
      string resp = SendRequest("POST", url, data, headers);;
      if(m_verbose)
        {
         SaveToFile(filename + "_post.txt", data);
         Print(resp);
        }
      return (getTokenValue(resp, "media_id"));
     }

  };
//+------------------------------------------------------------------+
